#-------------------------------------------------------------------------------
# Name:        Create Density Dots
# Purpose:     Create smaller number of points each representing
#              a fixed number of points in a larger point feature class
#              To facilitate dot-density mapping from point data
# Author:      Barry Joel Kronenfeld
# Created:     February 2014
# Copyright:   (c) bjkronenfeld 2014
# Licence:     GNU Public License v3
#-------------------------------------------------------------------------------

import arcpy
import arcpy.da
import numpy
import random
arcpy.env.overwriteOutput=True
arcpy.env.addOutputsToMap=True

def showMsg(msg):
    arcpy.AddMessage(msg)
    print msg

def getLicenses():
    arcpy.env.overwriteOutput=True
    # Check out Extensions
    # Spatial Analyst
    showMsg("Checking out licenses...")

    GotLicenses=True
    if arcpy.CheckExtension("Spatial") == "Available":
        arcpy.CheckOutExtension("Spatial")
        # Geostatistical Analyst
        if arcpy.CheckExtension("GeoStats") == "Available":
            arcpy.CheckOutExtension("GeoStats")
        else:
            print "Geostats License Error"
            GotLicenses=False
    else:
        print "Spatial Analyst license error"
        GotLicenses=False
    return GotLicenses

def getDensitySurface(inPtLyr,cellsAcross, popField="None", cellRadius=1):
    pass
    # Get extent of input points layer
    desc=arcpy.Describe(inPtLyr)
    inExt=desc.extent
    # note that extent may have different coordinate system than layer!!!
    sr_layer=desc.spatialReference # coordinate system of layer
    sr_map=inExt.spatialReference # coordinate system of the map
    arcpy.env.outputCoordinateSystem=sr_map
    # get min/max values
    xmin=inExt.XMin
    xmax=inExt.XMax
    ymin=inExt.YMin
    ymax=inExt.YMax
    # Get point density raster
    showMsg("Creating density raster...")
    xDiff=xmax-xmin
    yDiff=ymax-ymin
    minDiff = min(xDiff,yDiff)
    outCellSize=minDiff/cellsAcross
    nbrhood = arcpy.sa.NbrCircle(cellRadius,"CELL")
    densitySurface=arcpy.sa.PointDensity(inPtLyr,popField,outCellSize,nbrhood)

    # Calculate statistics on raster - necessary to avoid errors
    showMsg("Calculating raster statistics...")
    arcpy.CalculateStatistics_management(densitySurface)
    # Rescale so maximum is one
    showMsg("Rescaling raster...")
    # before ensuring that the map extent and spatial reference are used,
    # the next two lines didn't work in ArcMap (but they worked from PyScripter!)
    densityProperties = arcpy.GetRasterProperties_management(densitySurface,"MAXIMUM")
    maxDensity=float(densityProperties.getOutput(0))
    # divide
    scaledDensitySurface=densitySurface/maxDensity
    # return result
    return scaledDensitySurface

def main():
    #-# START HERE
    arcpy.env.addOutputsToMap=False
    # CHECK LICENSES
    licenseCheck=getLicenses()
    if licenseCheck==True:
        # INITIALIZATION
        arcpy.Delete_management(r'in_memory')
        arcpy.env.workspace=r'in_memory'
        arcpy.env.scratchWorkspace=r'in_memory'
        # INPUT PARAMETERS
        # - inPtLyr: feature class containing original points, to be aggregated
        # - dotValue: number of original points per point in output
        # - optional cellsAcross: number of raster cells on shortest side, used to calculate density
        # - optional shiftLimitRatio: limits the distance that points can be displaced, as ratio
        #     of average nearest neighbor distance
        # - optional sparseAreaDampening: factor limiting displacement of points in sparse regions
        #     (0 to 1)
        # - clipLayer: polygon layer that limits the placement of density dots
        # - outPtLyr: output layer containing density dots
        showMsg("Reading inputs...")

##        inPtLyr=r'C:\Users\bjkronenfeld\Documents\Documents\Data\hurricanes\Haiyan 2013\PubData.gdb\BLDG_DD_test1'
##        dotValue=20
##        cellsAcross=500
##        shiftLimitRatio=1.6
##        sparseAreaDampening=0.6
##        clipLayer=r'C:\Users\bjkronenfeld\Documents\Documents\Data\hurricanes\Haiyan 2013\Phillipines.shp'
##        outPtLyr=r'C:\Users\bjkronenfeld\Documents\Documents\Data\hurricanes\Haiyan 2013\PubData.gdb\test'

        inPtLyr=arcpy.GetParameterAsText(0)
        dotValue=int(arcpy.GetParameterAsText(1))
        cellsAcross=int(arcpy.GetParameterAsText(2))
        clipLayer=arcpy.GetParameterAsText(3)
        outPtLyr=arcpy.GetParameterAsText(4)

        # DENSITY SURFACE
        popField="None"
        cellRadius=3
        # RANDOM SAMPLE POINTS
        # - get number of features in input
        numInput=int(arcpy.GetCount_management(inPtLyr).getOutput(0))
        # - Calculate number of features for output
        targetNum=float(numInput)/dotValue
        showMsg("Target: " + str(targetNum) + " random points")
        # - get random points
        showMsg("Iterating...")
        densitySurface=getDensitySurface(inPtLyr,cellsAcross, popField,cellRadius)
        randomPoints=arcpy.ga.CreateSpatiallyBalancedPoints(densitySurface,targetNum,r'in_memory\randomPoints')
        # clip them if clip layer is set
        clippedPoints=randomPoints
        if clipLayer != "":
            clippedPoints=arcpy.Clip_analysis(randomPoints,clipLayer)
        mergedPoints=clippedPoints
        numCreated=int(arcpy.GetCount_management(mergedPoints).getOutput(0))
        InflateBy=1.0
        gotEnough=False
        # adjust inflation factor until we get just the right number of points
        maxDotValueError=0.500001
        ptsPerDotValueError=((float(numInput)/dotValue)-float(numInput)/(dotValue+1))
        acceptPtDiff=maxDotValueError*ptsPerDotValueError
        showMsg("fine tuning...")
        numIts=1
        while abs(numCreated-targetNum)>acceptPtDiff:
            showMsg("This attempt: " + str(numCreated) + " points")
            numIts +=1
            if numCreated>targetNum:
                gotEnough=True
            if gotEnough==False and numIts % 5==0:
                cellsAcross = cellsAcross * 2
                cellRadius += 1
                densitySurface=getDensitySurface(inPtLyr,cellsAcross, popField,cellRadius)
                InflateBy=float(targetNum)/float(numCreated)
            else:
                InflateBy=InflateBy*float(targetNum)/float(numCreated)
            tryNumPts=targetNum*InflateBy
            randomPoints=arcpy.ga.CreateSpatiallyBalancedPoints(densitySurface,tryNumPts,r'in_memory\randomPoints')
            clippedPoints=arcpy.Clip_analysis(randomPoints,clipLayer)
            numCreated=int(arcpy.GetCount_management(clippedPoints).getOutput(0))
        outDotValue=float(numInput)/numCreated
        showMsg("Created " + str(numCreated) + " points")
        showMsg("Final Dot Value: " + str(outDotValue))
        # save to results
        arcpy.env.addOutputsToMap=True
        arcpy.CopyFeatures_management(clippedPoints,outPtLyr)
        # CLEAN UP
        # - Release what needs to be released
        arcpy.CheckInExtension("Spatial")
        arcpy.CheckInExtension("GeoStats")
        arcpy.Delete_management(r'in_memory')
        # add output to map
        mxd=arcpy.mapping.MapDocument("CURRENT")
        df=arcpy.mapping.ListDataFrames(mxd,"*")[0]
        resultlyr=arcpy.mapping.Layer(outPtLyr)
        arcpy.mapping.AddLayer(df,resultlyr,"TOP")
        arcpy.RefreshActiveView()
        arcpy.RefreshTOC()
        arcpy.RefreshCatalog(outPtLyr)
        # - report success
        showMsg("Finished successfully.")
        showMsg("Results saved in " + outPtLyr)


    else: # failed to check out license
        showMsg("License error: Spatial Analyst or Geostatistical Analyst not available.")

if __name__ == "__main__":
    main()