#-------------------------------------------------------------------------------
# Name:        Spread Points
# Purpose:     Spread out points to avoid visual overlap.
# Author:      Barry Joel Kronenfeld
# Created:     February 2014
# Copyright:   (c) bjkronenfeld 2014
# Licence:     GNU Public License v3
#-------------------------------------------------------------------------------

import arcpy
import arcpy.da
import numpy
import random
arcpy.env.overwriteOutput=True
arcpy.env.addOutputsToMap=True

def showMsg(msg):
    arcpy.AddMessage(msg)
    print msg

def getLicenses():
    arcpy.env.overwriteOutput=True
    # Check out Extensions
    # Spatial Analyst
    showMsg("Checking out licenses...")

    GotLicenses=True
    if arcpy.CheckExtension("Spatial") == "Available":
        arcpy.CheckOutExtension("Spatial")
        # Geostatistical Analyst
        if arcpy.CheckExtension("GeoStats") == "Available":
            arcpy.CheckOutExtension("GeoStats")
        else:
            print "Geostats License Error"
            GotLicenses=False
    else:
        print "Spatial Analyst license error"
        GotLicenses=False
    return GotLicenses


# function to estimate the average distance
def averageTerritorialDistance(ptFC, clipFC="None"):
    # if clipFC is nothing, get convex hull around points
    if clipFC=="None":
        clipFC=arcpy.MinimumBoundingGeometry_management(ptFC,geometry_type="CONVEX_HULL")
    # get area of clip feature class
    clipCursor=arcpy.da.SearchCursor(clipFC,["SHAPE@AREA"])
    totArea=0
    for row in clipCursor:
        totArea += row[0]
    del row
    del clipCursor
    # divide by number of points
    countOut=arcpy.GetCount_management(ptFC)
    numPts=int(countOut.getOutput(0))
    areaPerPt=totArea/numPts
    # get radius, assuming circle
    areaByPI=areaPerPt/numpy.pi
    # get distance
    return numpy.sqrt(areaByPI)

def shiftPointsRandomly(Input_Features, shiftAmount=0.01):
    # randomly shifts input features by designated amount
    # to avoid TIN errors
    g=arcpy.CopyFeatures_management(Input_Features,arcpy.Geometry())
    ud=arcpy.da.UpdateCursor(Input_Features,['SHAPE'])
    curID=0
    for row in ud:
        # get X and Y values
        geo=g[curID]
        x=geo.centroid.X
        y=geo.centroid.Y
        # get distance moved
        xshift=(random.random()-0.5)*shiftAmount
        yshift=(random.random()-0.5)*shiftAmount
        x=x+xshift
        y=y+yshift
        # update
        point = arcpy.Point()
        point.X=x
        point.Y=y
        row[0]=point
        ud.updateRow(row)
        # move to next point
        curID+=1
    del row
    del ud

def diffusePoints(Input_Features, Clip_Features, numIterations=5, shiftLimitRatio=1, sparseRegionStability=0.5):
    # Report Start
    showMsg("Diffusing points...")
    numInput=int(arcpy.GetCount_management(Input_Features).getOutput(0))
    numClip=int(arcpy.GetCount_management(Clip_Features).getOutput(0))
    print str(numInput)+" points to process..."
    print str(numClip) + " polygons in clip feature class..."
    # Local variables:
    numIts = int(numIterations)
    # get base distance (maximum distance a point can be moved per iteration)
    print "getting base distance..."
    nnd=averageTerritorialDistance(Input_Features, Clip_Features)
    nnd=nnd*2
    baseD=shiftLimitRatio*nnd
    dblNumIts=float(numIts)
    baseD=baseD/dblNumIts
    pointMovePct=[0.0]*numInput
    # copy points
    print "copying points..."
    firstPointFC = r'in_memory\bob' + str(0)
    arcpy.CopyFeatures_management(Input_Features,firstPointFC)
    # do iterations
    for i in range(1,numIts+1):
        # Process: Create Thiessen Polygons
        print "creating Thiessen polygons..."
        lastPointFC= r'in_memory\bob' + str(i-1)
        thisTPoly=r'in_memory\T' + str(i)
        try:
            arcpy.CreateThiessenPolygons_analysis(lastPointFC, thisTPoly, "ONLY_FID")
        except Exception as e:
            showMsg("TIN Error: aborting and returning current points.")
            return lastPointFC
        # Process: Clip
        print "clipping..."
        thisTPolyClip=r'in_memory\TClip' + str(i)
        arcpy.Clip_analysis(thisTPoly, Clip_Features, thisTPolyClip, "")
        # Process: Feature To Point
        print "extracting centroids..."
        thisPointFC = r'in_memory\bob' + str(i)
        arcpy.FeatureToPoint_management(thisTPolyClip, thisPointFC, "CENTROID")
        # limit movement (?)
        # get last point FC as list of geometries
        print "getting geometries..."
        g1=arcpy.CopyFeatures_management(lastPointFC,arcpy.Geometry())
        thisPtSorted=arcpy.Sort_management(thisPointFC,r'in_memory\sorted'+str(i),'Input_FID')
        g2 = arcpy.CopyFeatures_management(thisPtSorted,arcpy.Geometry())
        # get areas, move percentage (first time around only)
        numPts=len(g1)
        area=[0.0]*numPts
        ThiessenSorted=arcpy.Sort_management(thisTPolyClip,r'in_memory\TSorted','Input_FID')
        gThiessen=arcpy.CopyFeatures_management(ThiessenSorted,arcpy.Geometry())
        numPts=int(arcpy.GetCount_management(gThiessen).getOutput(0))
        if i==1:
            maxArea=0
            # get maximum area
            for geoNum in range(0,numPts):
                area[geoNum]=gThiessen[geoNum].area
                if area[geoNum]>maxArea: maxArea=area[geoNum]
            # get area, point move percent
            for geoNum in range(0,numPts):
                pointMovePct[geoNum]=(1-sparseRegionStability) + area[geoNum]/maxArea*(sparseRegionStability)
        # loop through points
        print "adjusting geometries..."
        ud=arcpy.da.UpdateCursor(thisPointFC,['SHAPE'])
        curID=0
        for row in ud:
            # get X and Y values
            geo1=g1[curID]
            geo2=g2[curID]
            x1=geo1.centroid.X
            y1=geo1.centroid.Y
            x2=geo2.centroid.X
            y2=geo2.centroid.Y
            # get distance moved
            d=((x1-x2)**2)+((y1-y2)**2)
            d=d**0.5
            # initialize new coordinates
            newX2=x1
            newY2=y1
            # adjust coordinate (truncate shift)
            if d>0:
                newD=(d*baseD)/(d+baseD)
                thisMovePct=newD/d
                thisMovePct=thisMovePct*pointMovePct[curID]
                newX2=x1+(x2-x1)*thisMovePct
                newY2=y1+(y2-y1)*thisMovePct
            # update
            point = arcpy.Point()
            point.X=newX2
            point.Y=newY2
            row[0]=point
            ud.updateRow(row)
            # move to next point
            curID+=1
        del row
        del ud
        # report progress
        msg= "completed iteration " + str(i) + " of " + str(numIts)
        showMsg(msg)
    #return result of last iteration
    return thisPointFC

def getDensitySurface(inPtLyr,cellsAcross, popField="None"):
    pass
    # Get extent of input points layer
    desc=arcpy.Describe(inPtLyr)
    inExt=desc.extent
    # note that extent may have different coordinate system than layer!!!
    sr_layer=desc.spatialReference # coordinate system of layer
    sr_map=inExt.spatialReference # coordinate system of the map
    arcpy.env.outputCoordinateSystem=sr_map
    # get min/max values
    xmin=inExt.XMin
    xmax=inExt.XMax
    ymin=inExt.YMin
    ymax=inExt.YMax
    # Get point density raster
    showMsg("Creating density raster...")
    xDiff=xmax-xmin
    yDiff=ymax-ymin
    minDiff = min(xDiff,yDiff)
    outCellSize=minDiff/cellsAcross
    radius=outCellSize*5
    nbrCirc = arcpy.sa.NbrCircle(radius,"MAP")
    densitySurface=arcpy.sa.PointDensity(inPtLyr,popField,outCellSize)

    # Calculate statistics on raster - necessary to avoid errors
    showMsg("Calculating raster statistics...")
    arcpy.CalculateStatistics_management(densitySurface)
    # Rescale so maximum is one
    showMsg("Rescaling raster...")
    # before ensuring that the map extent and spatial reference are used,
    # the next two lines didn't work in ArcMap (but they worked from PyScripter!)
    densityProperties = arcpy.GetRasterProperties_management(densitySurface,"MAXIMUM")
    maxDensity=float(densityProperties.getOutput(0))
    # divide
    scaledDensitySurface=densitySurface/maxDensity
    # return result
    return scaledDensitySurface

def getPositiveDefiniteRegion(densityRaster, threshold=0):
    # returns a polygon feature class representing the region
    # where the values of the density raster are greater than threshold
    # use condition tool
    import arcpy.sa
    outcon= arcpy.sa.Con(densityRaster,1,"","Value > " + str(threshold))
    outpoly=arcpy.RasterToPolygon_conversion(outcon)
    return outpoly

def main():
    #-# START HERE
    arcpy.env.addOutputsToMap=False
    # CHECK LICENSES
    licenseCheck=getLicenses()
    if licenseCheck==True:
        # INITIALIZATION
        arcpy.Delete_management(r'in_memory')
        arcpy.env.workspace=r'in_memory'
        #arcpy.env.scratchWorkspace=r'in_memory'
        # INPUT PARAMETERS
        # - inPtLyr: feature class containing original points, to be aggregated
        # - numIterations: number of iterations to apply
        # - optional shiftLimitRatio: limits the distance that points can be displaced, as ratio
        #     of average nearest neighbor distance
        # - optional sparseRegionStability: factor limiting displacement of points in sparse regions
        #     (0 to 1)
        # - clipLayer: polygon layer that limits the placement of density dots
        # - outPtLyr: output layer containing density dots
        showMsg("Reading inputs...")

##        inPtLyr=r'C:\Users\bjkronenfeld\Documents\Documents\Data\hurricanes\Haiyan 2013\PubData.gdb\BLDG_DD_test1'
##        numIterations=5
##        shiftLimitRatio=1.6
##        sparseRegionStability=0.6
##        clipLayer=r'C:\Users\bjkronenfeld\Documents\Documents\Data\hurricanes\Haiyan 2013\Phillipines.shp'
##        outPtLyr=r'C:\Users\bjkronenfeld\Documents\Documents\Data\hurricanes\Haiyan 2013\PubData.gdb\test'

        inPtLyr=arcpy.GetParameterAsText(0)
        numIterations=int(arcpy.GetParameterAsText(1))
        shiftLimitRatio=float(arcpy.GetParameterAsText(2))
        sparseRegionStability=float(arcpy.GetParameterAsText(3))
        clipLayer=arcpy.GetParameterAsText(4)
        outPtLyr=arcpy.GetParameterAsText(5)
        # make copy of input points
        inCopy=arcpy.CopyFeatures_management(inPtLyr, r'in_memory\inCopy')
        # - move by a random amount to avoid TIN errors
        shiftPointsRandomly(inCopy,0.01)
        # get buffer around input
        print "buffering..."
        popField="None"
        densitySurface=getDensitySurface(inPtLyr,500, popField)
        buffPoly=getPositiveDefiniteRegion(densitySurface)
        if clipLayer<>"":
            showMsg("clipping data...")
            inCopy=arcpy.Clip_analysis(inCopy,clipLayer)
            buffPoly=arcpy.Clip_analysis(buffPoly,clipLayer)
        # diffuse points
        diffusedPoints= diffusePoints(inCopy,buffPoly,numIterations,shiftLimitRatio,sparseRegionStability)
        # save to results
        arcpy.env.addOutputsToMap=True
        arcpy.CopyFeatures_management(diffusedPoints,outPtLyr)
        # CLEAN UP
        # - Release what needs to be released
        arcpy.CheckInExtension("Spatial")
        arcpy.CheckInExtension("GeoStats")
        arcpy.Delete_management(r'in_memory')
        # add output to map
        mxd=arcpy.mapping.MapDocument("CURRENT")
        df=arcpy.mapping.ListDataFrames(mxd,"*")[0]
        resultlyr=arcpy.mapping.Layer(outPtLyr)
        arcpy.mapping.AddLayer(df,resultlyr,"TOP")
        arcpy.RefreshActiveView()
        arcpy.RefreshTOC()
        arcpy.RefreshCatalog(outPtLyr)
        # - report success
        showMsg("Finished successfully.")
        showMsg("Results saved in " + outPtLyr)


    else: # failed to check out license
        showMsg("License error: Spatial Analyst or Geostatistical Analyst not available.")

if __name__ == "__main__":
    main()